from .models.lge.spec_l import ModelScraper_l as Specscraper_l
from .models.sony.spec_s import ModelScraper_s as Specscraper_s
from .models.sony.sepc_sjp import ModelScraper_sjp as Specscraper_sjp
from .models.pana.spec_p import ModelScraper_p as Specscraper_p
from .models.pana.spec_pjp import ModelScraper_pjp as Specscraper_pjp

from market_research.scraper.models.sony.visualizer_s import Visualizer_s
from .rtings.rtings import Rtings
from .rtings.rvisualizer import Rvisualizer
from .rtings.rurlsearcher import Rurlsearcher


from bs4 import BeautifulSoup
import requests
import time
import pandas as pd
from tqdm import tqdm
import re
from selenium.webdriver.common.by import By
from tools.file import FileManager
from market_research.scraper._scaper_scheme import Scraper




class ModelScraper_l(Scraper):
    def __init__(self, enable_headless=True,
                 export_prefix="lge_model_info_web", intput_folder_path="input", output_folder_path="results",
                 verbose: bool = False, wait_time=2):

        super().__init__(enable_headless=enable_headless, export_prefix=export_prefix,
                         intput_folder_path=intput_folder_path, output_folder_path=output_folder_path)

        self.tracking_log = verbose
        self.wait_time = wait_time
        self.file_manager = FileManager
        self.log_dir = "logs/lge/models"
        if self.tracking_log:
            FileManager.make_dir(self.log_dir)

    def get_models_info(self, format_df: bool = True, fastmode: bool = False, show_visit:bool=False):
        print("collecting models")
        url_series_set = self._get_url_series()
        url_series_dict = {}
        for url in url_series_set:
            url_models = self._get_models(url=url)
            url_series_dict.update(url_models)
        print("number of total model:", len(url_series_dict))
        if fastmode:
            model_list = list(url_series_dict.keys())
            return model_list
        print("collecting spec")
        visit_url_dict = {}
        dict_models = {}
        cnt_loop=2
        for cnt in range(cnt_loop):#main try
            for key, url_model in tqdm(url_series_dict.items()):
                try:
                    dict_info = self._get_model_info(url_model)
                    dict_models[key] = dict_info
                    dict_spec = self._get_global_spec(url=url_model)
                    dict_models[key].update(dict_spec)
                    visit_url_dict[key] = url_model
                except Exception as e:
                    if cnt == cnt_loop - 1 :
                        print(f"\nFailed to get info from {key}")
                        print(e)
                    pass
            break
        if show_visit:
            print("\n")
            for model, url in visit_url_dict.items():  print(f"{model}: {url}")

        if format_df:
            df_models = pd.DataFrame.from_dict(dict_models).T
            FileManager.df_to_excel(df_models.reset_index(), file_name=self.output_xlsx_name, sheet_name="raw_na",
                                    mode='w')
            return df_models
        else:
            return dict_models

    def _get_url_series(self) -> set:
        """
        Get the series URLs by scrolling down the main page.
        """
        url = "https://www.lg.com/us/tvs"
        prefix = "https://www.lg.com/"
        step = 200
        url_series = set()
        try_total = 5
        for _ in range(2): #page_checker
            for _ in range(try_total):
                driver = self.web_driver.get_chrome()
                try:
                    driver.get(url=url)
                    time.sleep(self.wait_time)
                    scroll_distance_total = self.web_driver.get_scroll_distance_total()
                    scroll_distance = 0

                    while scroll_distance < scroll_distance_total:
                        for _ in range(2):
                            html = driver.page_source
                            soup = BeautifulSoup(html, 'html.parser')
                            elements = soup.find_all('a', class_="css-11xg6yi")
                            for element in elements:
                                url_series.add(prefix + element['href'].strip())
                            driver.execute_script(f"window.scrollBy(0, {step});")
                            time.sleep(self.wait_time)
                            scroll_distance += step
                    driver.quit()
                    break
                except Exception as e:
                    driver.quit()
                    if self.tracking_log:
                        print(f"Try collecting {_ + 1}/{try_total}")
                        print(e)
        print("The website scan has been completed.")
        print(f"number of total series: {len(url_series)}")
        return url_series

    def _get_models(self, url: str, prefix="https://www.lg.com/", static_mode=True) -> dict:
        """
        Extract all model URLs from a given series URL.
        """
        try_total = 5
        for cnt_try in range(try_total):
            try:
                dict_url_models = {}
                for _ in range(3):
                    if static_mode:
                        response = requests.get(url)
                        page_content = response.text
                    else:
                        driver = self.web_driver.get_chrome()
                        driver.get(url=url)
                        time.sleep(self.wait_time)
                        page_content = driver.page_source
                    soup = BeautifulSoup(page_content, 'html.parser')
                    elements = soup.find_all('a', class_='css-1a0ki8h')

                    for element in elements:
                        try:
                            element_url = prefix + element['href']
                            label = self.file_manager.get_name_from_url(element_url)
                            dict_url_models[label] = element_url.strip()
                        except Exception as e:
                            print(f"Getting series error ({e})")
                            pass
                if self.tracking_log:
                    print(f"LGE {self.file_manager.get_name_from_url(url)[4:]} series: {len(dict_url_models)}")
                for key, value in dict_url_models.items():
                    if self.tracking_log:
                        print(f'{key}: {value}')
                return dict_url_models
            except Exception as e:
                if self.tracking_log:
                    print(f"_get_models try: {cnt_try + 1}/{try_total}")

    def _get_model_info(self, url: str) -> dict:
        """
        Extract model information (name, price, description) from a given model URL.
        """
        response = requests.get(url)
        if self.tracking_log:
            print(" Connecting to", url)
        page_content = response.text
        soup = BeautifulSoup(page_content, 'html.parser')
        dict_info = {}
        label = soup.find('span', class_="MuiTypography-root MuiTypography-overline css-rmroyy").text.strip()
        dict_info["model"] = label.split()[-1]
        try:
            pattern = r'\$(\d+,\d+\.\d{2})'  # 달러 기호 후 숫자, 콤마, 소수점 및 두 자리 숫자 추출
            price = soup.find('div', class_='MuiBox-root css-ayrjh9').text.strip()

            matches = re.findall(pattern, price)
            if len(matches) == 3:
                dict_info["price"] = f"{matches[0]} {matches[1]}"
            else:
                dict_info["price"] = price
        except:
            dict_info["price"] = None
        # dict_info.update(self._extract_model_info(dict_info.get("model")))
        try:
            dict_info["description"] = soup.find('h2',
                                                 class_='MuiTypography-root MuiTypography-h5 css-ff7u73').text.strip()
        except:
            try:
                dict_info["description"] = soup.find('h1',
                                                     class_='MuiTypography-root MuiTypography-h5 css-ff7u73').text.strip()
            except:
                dict_info["description"] = ""
        return dict_info

    def _get_global_spec(self, url: str, ) -> dict:
        """
        Extract global specifications from a given model URL.
        """
        try_total = 1
        model = None
        driver = None
        case = 0
        for cnt_try in range(try_total):
            try:
                driver = self.web_driver.get_chrome()
                driver.get(url=url)
                model = self.file_manager.get_name_from_url(url)
                dir_model = f"{self.log_dir}/{model}"
                stamp_today = self.file_manager.get_datetime_info(include_time=False)
                stamp_url = self.file_manager.get_name_from_url(url)
                if self.tracking_log:
                    self.file_manager.make_dir(dir_model)
                if self.tracking_log:
                    driver.save_screenshot(f"./{dir_model}/{stamp_url}_0_model_{stamp_today}.png")
                time.sleep(self.wait_time)

                for id in range(5):
                    element_all_specs = driver.find_element(By.ID, f"simple-tab-{id}")
                    if element_all_specs.text == "ALL SPECS":
                        case = id
                        break

                self.web_driver.move_element_to_center(element_all_specs)
                time.sleep(self.wait_time)
                element_all_specs.click()
                element_click_expand_all = driver.find_element(By.CSS_SELECTOR,
                                                               ".MuiTypography-root.MuiTypography-h6.MuiTypography-alignLeft.MuiLink-root.MuiLink-underlineAlways.css-m4r0k4")
                self.web_driver.move_element_to_center(element_click_expand_all)
                element_click_expand_all.click()

                time.sleep(self.wait_time)

                if self.tracking_log:
                    driver.save_screenshot(
                        f"./{dir_model}/{stamp_url}_1_element_all_specs_{stamp_today}.png")
                dict_spec = dict()
                spec_table_element = driver.find_element(By.XPATH, f'//*[@id="simple-tabpanel-{case}"]')
                spec_elements = spec_table_element.find_elements(By.CSS_SELECTOR, '.MuiBox-root.css-1nnt9ji')

                for spec_element in spec_elements:
                    label = spec_element.find_element(By.CSS_SELECTOR,
                                                      '.MuiTypography-root.MuiTypography-caption.css-1k7x5zg').text.strip()
                    content = spec_element.find_element(By.CSS_SELECTOR,
                                                        '.MuiTypography-root.MuiTypography-body2.css-10wyomj').text.strip()
                    if label != '':
                        # if self.tracking_log:
                        #     print(f"{label}: {content}")
                        dict_spec[label] = content

                dict_spec.pop("*", None)

                driver.quit()
                if self.tracking_log:
                    print(f"Received information from {url}")
                return dict_spec
            except Exception as e:
                if self.tracking_log:
                    print(f"An error occurred on page 3rd : {model} try {cnt_try + 1}/{try_total}")
                    print(e)
                driver.quit()
                pass

    def _extract_model_info(self, model: str):
        """
        소문자로 시작하는 알파벳('type'), 뒤에 숫자('size'), 그 다음에 소문자 알파벳(시리즈)과 숫자(연도)가 섞인 'model',
        마지막에 숫자가 포함되지 않는 문자열(소문자)을 'grade'로 추출
        """
        dict_info = {}
        pattern = r'([a-z]+)(\d+)([a-z]+)(\d+)([a-z]+)$'
        match = re.search(pattern, model.lower())
        if match:
            # 그룹화된 각 부분을 추출
            dict_info["type"] = match.group(1)  # 예: 'oled'
            dict_info["size"] = match.group(2)  # 예: '65'
            dict_info["series"] = match.group(3)  # 예: 'g'
            dict_info["year"] = match.group(4)  # 예: '4'
            dict_info["grade"] = match.group(5)  # 예: 'pus'

        year_mapping = {
            '1': "2021",
            '2': "2022",
            '3': "2023",
            '4': "2024",
            '5': "2024",
            '6': "2024",
            # Add additional mappings as needed
        }
        try:
            dict_info["year"] = year_mapping.get(dict_info.get("year"))
        except:
            dict_info["year"] = None
        return dict_info

    def _soup_to_dict(self, soup):
        """
        Convert BeautifulSoup soup to dictionary.
        """
        try:
            h4_tag = soup.find('h4').text.strip()
            p_tag = soup.find('p').text.strip()
        except:
            try:
                h4_tag = soup.find('h4').text.strip()
                p_tag = ""
            except Exception as e:
                print("Parser error", e)
                h4_tag = "Parser error"
                p_tag = "Parser error"
        return {h4_tag: p_tag}



import time
from selenium.webdriver.common.by import By
from bs4 import BeautifulSoup
import pandas as pd
import re
from tqdm import tqdm
from tools.file import FileManager
from market_research.scraper._scaper_scheme import Scraper

class Rtings(Scraper):
    def __init__(self, enable_headless = True,
                         export_prefix= "rtings",
                         intput_folder_path = "input",
                         output_folder_path = "results", verbose:bool=False, wait_time=1):

        super().__init__(enable_headless = enable_headless,
                         export_prefix= export_prefix,
                         intput_folder_path = intput_folder_path,
                         output_folder_path = output_folder_path)
                             
        self.verbose = verbose
        self.wait_time = wait_time
    def get_data(self, urls:list, export_excel=True):
        score_df= pd.DataFrame()
        measurement_df = pd.DataFrame()
        comments_df = pd.DataFrame()
        url = None
        try:
            for url in tqdm(urls):
                if self.verbose:
                    print(f"connecting to {url}")
                df = self._get_score(url, format_df=True)
                score_df = pd.concat([score_df, df], axis=0)

                # 저장할 데이터 경로
                df = self._get_measurement_reuslts(url)
                measurement_df = pd.concat([measurement_df, df], axis=0)

                df = self._get_commetns(url, format_df=True)
                comments_df = pd.concat([comments_df, df], axis=0)

            if export_excel:
                FileManager.df_to_excel(score_df, file_name=self.output_xlsx_name, sheet_name="scores", mode='w')
                FileManager.df_to_excel(measurement_df, file_name=self.output_xlsx_name, sheet_name="measurement", mode='a')
                FileManager.df_to_excel(comments_df, file_name=self.output_xlsx_name, sheet_name="comments", mode='a')
        except Exception as e:
            if self.verbose:
                print(f"fail {url}")
                print(e)
        return {
            "scores":score_df,
                "measurement":measurement_df,
                "comments":comments_df
        }
        

    def _get_score(self, url:str="https://www.rtings.com/tv/reviews/sony/a95l-oled", format_df=True) :
        """
        return type
        -> dict|pd.DataFrame
        """

        driver = self.web_driver.get_chrome()
        maker = url.split("/")[-2]
        model = url.split("/")[-1]
        url = url.lower()
        # 웹 페이지 로드
        driver.get(url)
        time.sleep(self.wait_time)

        # scorecard-row-content 클래스를 가진 요소들을 선택
        elements = driver.find_elements(By.CLASS_NAME, "scorecard-row-content")
        maker_dict = {}
        try:
            score_dict = {}
            for element in elements:
                label = element.find_element(By.CLASS_NAME, 'scorecard-row-name').text.strip()
                score = element.find_element(By.CLASS_NAME, 'e-score_box-value ').text.strip()
                score_dict[label] = score

            # WebDriver 종료
            driver.quit()

            model_dict = {model:score_dict}
            maker_dict = {maker:model_dict}

            rows = []
            # 딕셔너리를 순회하며 데이터프레임 행으로 추가
            for maker, models in maker_dict.items():
                for model, scores in models.items():
                    for score_type, score in scores.items():
                        rows.append({'Maker': maker, 'Model': model, 'Score Type': score_type, 'Score': score})
            scores_df = pd.DataFrame(rows)
            scores_dict = scores_df.to_dict()
            if format_df:
                return scores_df
            else:
                return scores_dict

        except Exception as e:
            if self.verbose:
                print(f"get specification error: {e}")
            # WebDriver 종료
            driver.quit()



    def _get_commetns(self, url:str="https://www.rtings.com/tv/reviews/sony/a95l-oled", format_df=True, min_sentence_length = 0):
        """
        return dict or DataFrame

        """

        # Selenium을 사용하여 웹 드라이버 시작
        driver = self.web_driver.get_chrome()
        maker = url.split("/")[-2]
        product = url.split("/")[-1]
        url = url.lower()
        # url = url +"#page-comments"
        driver.get(url)
        time.sleep(self.wait_time)
        ## Load More 버튼 열기
        try:
            while True:
                try:
                    button_div = driver.find_element(By.CLASS_NAME, "comment_list-footer")
                    button = button_div.find_element(By.CLASS_NAME, "e-button")
                    button.click()
                    time.sleep(1)
                except:
                    break
            try:
                page_source = driver.page_source
                soup = BeautifulSoup(page_source, 'html.parser')
                comment_contents = soup.find_all('div', class_='comment_list-item-content e-discussion_content is-newest')
                comments_list = []

                for idx, comment_content in enumerate(comment_contents):
                    quote_controls = comment_content.find('div', class_='quote-controls')
                    if quote_controls:
                        quote_controls.decompose()
                    quote_content = comment_content.find('div', class_='quote-content')
                    if quote_content:
                        quote_content.decompose()
                    comment_text = comment_content.get_text(strip=True, separator='\n')
                    comment_text = re.sub(r'https?://\S+', '', comment_text)

                    # min_sentence_length 길이 이상의 문장만 분석
                    if len(comment_text.split()) >= min_sentence_length:
                        comments_list.append(
                            {'idx': idx, 'maker': maker, 'product': product, 'sentences': comment_text})
                comments_df = pd.DataFrame(comments_list).set_index("idx")
                comments_dict = comments_df.to_dict()
            finally:
                driver.quit()
        except:
            if self.verbose:
                print("no comment")
            comments_df = pd.DataFrame([{'idx': 0, 'maker': maker, 'product': product, 'sentences': "No data"}]).set_index("idx")
            comments_dict = comments_df.to_dict()

        if format_df:
            return comments_df
        else:
            return comments_dict

    def _get_measurement_reuslts(self, url: str = "https://www.rtings.com/tv/reviews/sony/a95l-oled") ->pd.DataFrame:

        driver = self.web_driver.get_chrome()
        maker = url.split("/")[-2]
        product = url.split("/")[-1]
        url = url.lower()
        driver.get(url)
        time.sleep(self.wait_time)
        page_source = driver.page_source
        driver.quit()
        
        soup = BeautifulSoup(page_source, 'html.parser')
        
    

        # # product_page-category 텍스트 찾기
        # product_category = test_group.find('a', class_='e-anchor').get_text(strip=True)
        # print(f"product_category: {product_category}")
        # 딕셔너리 초기화
        results_list = []
        category = ""

        test_groups = soup.find_all('div', class_='test_group e-simple_grid-item-2-pad')
        for test_group in test_groups:
            # test_group-category 텍스트 찾기
            # category = test_group.find('div', class_='test_group-category').get_text(strip=True)
            header_element = test_group.find('div', class_='test_group-header')
            category_element = header_element.find('div', class_='test_group-category')
            if category_element:
                category = category_element.get_text(strip=True)
                category_element.extract()
            # span 태그 내의 텍스트 추출 및 '_'로 구분
            scores_headder = '(><)_'.join(header_element.stripped_strings)
            # test_value is-number 찾기
            test_values = test_group.find_all('div', class_='test_value is-number')
            # 리스트 초기화
            for test_value in test_values:
                # test_value-label과 test_result_value 찾기
                label = test_value.find('span', class_='test_value-label').get_text(strip=True)
                result_value = test_value.find('span',
                                               class_='test_result_value e-test_result review-value-score').get_text(strip=True)
                # 딕셔너리에 추가
                results_list.append([maker, product, category, scores_headder, label, result_value])
            # 리스트를 데이터프레임으로 변환


        results_df = pd.DataFrame(results_list,
                                  columns=["maker", "product", "category", "scores_header", "label", "result_value"])
        # scores_header_list 생성 및 수정
        scores_header_list = results_df.scores_header.map(lambda x: x.split("_"))

        scores_header_list_s = []
        for scores_header in scores_header_list:
           if len(scores_header) <2:
               scores_header_list_s.append(["(><)", scores_header[0]])
           else:
               scores_header_list_s.append([scores_header[0], scores_header[1]])

        # print(scores_header_list_s)

           # scores_header_df 생성
        scores_header_df = pd.DataFrame(scores_header_list_s, columns=["score", "header"])
        scores_header_df["score"] = scores_header_df["score"].map(lambda x:x.replace("(><)", ""))
        # results_df와 scores_header_df 병합
        results_df = pd.concat([results_df, scores_header_df], axis=1)
        results_df = results_df[["maker", "product", "category", "header", "score", "label", "result_value"]]
        # 결과 확인

        # print(results_df.head())

        return results_df



from bs4 import BeautifulSoup
import pickle
import time
import pandas as pd
from tqdm import tqdm
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.common.by import By
from selenium.webdriver.common.action_chains import ActionChains
from tools.file import FileManager
from market_research.scraper._scaper_scheme import Scraper

class ModelScraper_s(Scraper):
    def __init__(self, enable_headless=True,
                 export_prefix="sony_model_info_web", intput_folder_path="input",  output_folder_path="results",
                 verbose: bool = False, wait_time=2):
        """
        Initialize the instance with the specified configuration.

        Parameters:
        enable_headless (bool): Whether to run the browser in headless mode. Default is True.
        export_prefix (str): Prefix for export file names. Default is "sony_model_info_web".
        intput_folder_path (str): Path to the input folder. Default is "input".
        output_folder_path (str): Path to the output folder. Default is "results".
        verbose (bool): If True, enables tracking logs. Default is False.
        wait_time (int): Time to wait for page load in seconds. Default is 1.

        If tracking_log is enabled:
        Deletes the existing log directory and creates a new one.
        """

        super().__init__(enable_headless=enable_headless, export_prefix=export_prefix,  intput_folder_path=intput_folder_path, output_folder_path=output_folder_path)

        self.tracking_log = verbose
        self.wait_time = wait_time
        self.file_manager = FileManager
        self.log_dir = "logs/sony/models"
        if self.tracking_log:
            FileManager.delete_dir(self.log_dir)
            FileManager.make_dir(self.log_dir)

    def get_models_info(self, format_df:bool=True, fastmode:bool=False, temporary_year_marking=False):
        """
        Collect model information from URLs and return the data in the desired format.

        Parameters:
        format_df (bool): If True, return the result as a DataFrame and save it to an Excel file. Default is True.
        fastmode (bool): If True, skips collecting global specifications for faster processing. Default is False.
        temporary_year_marking (bool): If True, temporarily mark the year as "2024" for models without a year. Default is False.

        Returns:
        pd.DataFrame or dict: A DataFrame of model information if format_df is True, otherwise a dictionary.
        """


        print("collecting models")
        url_series_set = self._get_url_series()

        url_series_dict = {}
        for url in url_series_set:
            url_models = self._get_models(url=url)
            url_series_dict.update(url_models)
            
        print("number of total model:", len(url_series_dict))
        print("collecting spec")
    
    
        if fastmode:
            print("operating fast mode")
        visit_url_dict = {}
        dict_models = {}
        for key, url_model in tqdm(url_series_dict.items()):
            dict_info = self._get_model_info(url_model)
            dict_models[key] = dict_info
            visit_url_dict[key] = url_model
            if fastmode is False:
                dict_spec = self._get_global_spec(url=url_model)
                dict_models[key].update(dict_spec)
        if self.tracking_log:
            print("\n")
            for model, url in visit_url_dict.items():  print(f"{model}: {url}")

        if format_df:
            df_models = pd.DataFrame.from_dict(dict_models).T
            if temporary_year_marking:
                df_models['year'] = df_models['year'].fillna("2024") ## 임시
            FileManager.df_to_excel(df_models.reset_index(), file_name=self.output_xlsx_name, sheet_name="raw_na", mode='w')
            return df_models
        else:
            return dict_models


    def _get_url_series(self) -> set:
        """
        Get the series URLs by scrolling down the main page.
        """
        url = "https://electronics.sony.com/tv-video/televisions/c/all-tvs/"
        prefix = "https://electronics.sony.com/"
        step = 200
        url_series = set()
        try_total = 5
        for _ in range(2): #page_checker
            for cnt_try in range(try_total):
                driver = self.web_driver.get_chrome()
                try:
                    driver.get(url=url)
                    time.sleep(self.wait_time)
                    scroll_distance_total = self.web_driver.get_scroll_distance_total()
                    scroll_distance = 0
                    while scroll_distance < scroll_distance_total:
                        for _ in range(2):
                            html = driver.page_source
                            soup = BeautifulSoup(html, 'html.parser')
                            elements = soup.find_all('a', class_="custom-product-grid-item__product-name")
                            for element in elements:
                                url_series.add(prefix + element['href'].strip())
                            driver.execute_script(f"window.scrollBy(0, {step});")
                            time.sleep(self.wait_time)
                            scroll_distance += step
                    break
                except Exception as e:
                    if self.tracking_log:
                        if cnt_try + 1 == try_total:
                            print(f"collecting primary url error from {url}")
                finally:
                    driver.quit()
        print("The website scan has been completed.")
        print(f"number of total series: {len(url_series)}")
        if self.tracking_log:
            print(url_series)
        return url_series

    def _get_models(self, url: str) -> dict:
        """
        Extract all model URLs from a given series URL.
        """
        model = self.file_manager.get_name_from_url(url)
        dir_model = f"{self.log_dir}/{model}"
        stamp_today = self.file_manager.get_datetime_info(include_time=False)
        stamp_url = self.file_manager.get_name_from_url(url)
        
        try_total = 5
        dict_url_models = {}
        for cnt_try in range(try_total):
            driver = self.web_driver.get_chrome()
            driver.get(url=url)
            time.sleep(self.wait_time)
            try:
                elements = driver.find_element(By.XPATH,
                                                      '//*[@id="PDPOveriewLink"]/div[1]/div/div/div[2]/div/app-custom-product-summary/div[2]/div/div[1]/app-custom-product-variants/div/app-custom-variant-selector/div/div[2]')

                url_elements = elements.find_elements(By.TAG_NAME, 'a')

                for url_element in url_elements:
                    url = url_element.get_attribute('href')
                    label = self.file_manager.get_name_from_url(url)
                    dict_url_models[label] = url.strip()
                break
            except Exception as e:
                if cnt_try + 1 == try_total:
                        print(f"Getting series error from {url}")
                        if self.tracking_log:
                            self.file_manager.make_dir(dir_model)
                            driver.save_screenshot(f"./{dir_model}/{stamp_url}_Getting series error_{stamp_today}.png")
            finally:
                driver.quit()

        if self.tracking_log:
            print(f"SONY {self.file_manager.get_name_from_url(url)[4:]} series: {len(dict_url_models)}")
        for key, value in dict_url_models.items():
            if self.tracking_log:
                print(f'{key}: {value}')
        return dict_url_models

    def _get_model_info(self, url: str) -> dict:
        """
        Extract model information (name, price, description) from a given model URL.
        """
        model = self.file_manager.get_name_from_url(url)
        dir_model = f"{self.log_dir}/{model}"
        stamp_today = self.file_manager.get_datetime_info(include_time=False)
        stamp_url = self.file_manager.get_name_from_url(url)
        try_total = 10
        dict_info = {
            "model": None,
            "description": None,
            "price": None,
            "price_original": None,
            "price_gap": None,
        }

        for cnt_try in range(try_total):
            driver = self.web_driver.get_chrome()
            try:
                driver.get(url)
                time.sleep(self.wait_time)
                
                # Extract model
                try:
                    model = driver.find_element(By.XPATH,
                                                '//*[@id="cx-main"]/app-product-details-page/div/app-custom-product-intro/div/div/div[1]/div/span').text

                    
                    model = model.split(":")[-1].strip()
                except Exception as e:                    
                    if self.tracking_log:
                        if cnt_try + 1 == try_total:
                            print(f"Model extraction failed from {url}")
                            self.file_manager.make_dir(dir_model)
                            driver.save_screenshot(f"./{dir_model}/{stamp_url}_Model extraction failed_{stamp_today}.png")
                    pass

                try:
                    description = driver.find_element(By.XPATH,
                                                      '//*[@id="cx-main"]/app-product-details-page/div/app-custom-product-intro/div/div/div[1]/h1/p').text
    
                except:
                    description = ""
                # Extract price
                try:
                    price_now = driver.find_element(By.XPATH,
                                                    '//*[@id="PDPOveriewLink"]/div[1]/div[2]/div[1]/div[2]/div/app-custom-product-summary/app-product-pricing/div/div[1]/p[1]').text
                    price_original = driver.find_element(By.XPATH,
                                                         '//*[@id="PDPOveriewLink"]/div[1]/div[2]/div[1]/div[2]/div/app-custom-product-summary/app-product-pricing/div/div[1]/p[2]').text
                    price_now = float(price_now.replace('$', '').replace(',', ''))
                    price_original = float(price_original.replace('$', '').replace(',', ''))
                    price_gap = price_original - price_now
                except:
                    try:
                        price_now = driver.find_element(By.XPATH,
                                                        '//*[@id="PDPOveriewLink"]/div[1]/div[2]/div[1]/div[2]/div/app-custom-product-summary/app-product-pricing/div/div[1]/p').text
                        price_now = float(price_now.replace('$', '').replace(',', ''))
                        price_original = price_now
                        price_gap = 0.0
                    except:
                        price_now = float('nan')
                        price_original = float('nan')
                        price_gap = float('nan')
                if self.tracking_log:
                    print(f"{model}: {description}")
                dict_info.update({
                    "model": model,
                    "description": description,
                    "price": price_now,
                    "price_original": price_original,
                    "price_gap": price_gap,
                })
                

                    
                dict_info.update(self._extract_model_info(dict_info.get("model")))
                break
            except Exception as e:
                pass
                if self.tracking_log:
                    print(f"error at get_model_info from{url}")
            finally:
                driver.quit()
        return dict_info

    def _get_global_spec(self, url: str) -> dict:
        """
        Extract global specifications from a given model URL.
        """
        try_total = 10
        driver = None
        if self.tracking_log:  print(" Connecting to", url)
        for cnt_try in range(try_total):
            try:
                dict_spec = {}
                driver = self.web_driver.get_chrome()
                driver.get(url=url)
                model = self.file_manager.get_name_from_url(url)
                dir_model = f"{self.log_dir}/{model}"
                stamp_today = self.file_manager.get_datetime_info(include_time=False)
                stamp_url = self.file_manager.get_name_from_url(url)

                if self.tracking_log:
                    self.file_manager.make_dir(dir_model)
                    driver.save_screenshot(f"./{dir_model}/{stamp_url}_0_model_{stamp_today}.png")
                time.sleep(self.wait_time)
                element_spec = driver.find_element(By.XPATH, '//*[@id="PDPSpecificationsLink"]')
                self.web_driver.move_element_to_center(element_spec)
                if self.tracking_log:
                    driver.save_screenshot(f"./{dir_model}/{stamp_url}_1_move_to_spec_{stamp_today}.png")
                time.sleep(self.wait_time)
                element_click_spec = driver.find_element(By.XPATH, '//*[@id="PDPSpecificationsLink"]/cx-icon')
                element_click_spec.click()
                time.sleep(self.wait_time)
                if self.tracking_log:
                    driver.save_screenshot(
                        f"./{dir_model}/{stamp_url}_2_after_click_specification_{stamp_today}.png")
                    
                    
                for _ in range(5):
                    try:
                        # 팝업 닫기 버튼을 기다렸다가 클릭
                        close_button = driver.find_element(By.XPATH, '//*[@id="contentfulModalClose"]')
                        close_button.click()
                        break
                    except Exception as e:
                        pass
                    
                        
                try:
                    element_see_more = driver.find_element(By.XPATH,'//*[@id="cx-main"]/app-product-details-page/div/app-product-specification/div/div[2]/div[3]/button')
                    self.web_driver.move_element_to_center(element_see_more)
                    if self.tracking_log:
                        driver.save_screenshot(f"./{dir_model}/{stamp_url}_3_after_click_see_more_{stamp_today}.png")
                    element_see_more.click()
                except:
                    try:
                        element_see_more = driver.find_element(By.XPATH,
                                                               '//*[@id="cx-main"]/app-product-details-page/div/app-product-specification/div/div[2]/div[2]/button')
                        self.web_driver.move_element_to_center(element_see_more)
                        if self.tracking_log:
                            driver.save_screenshot(
                                f"./{dir_model}/{stamp_url}_3_after_click_see_more_{stamp_today}.png")
                        element_see_more.click()
                    except:
                        if self.tracking_log:
                            print("Cannot find the 'see more' button on the page")
                time.sleep(self.wait_time)
                driver.find_element(By.ID, "ngb-nav-0-panel").click()
                for _ in range(15):
                    elements = driver.find_elements(By.CLASS_NAME,"full-specifications__specifications-single-card__sub-list")
                    for element in elements:
                        soup = BeautifulSoup(element.get_attribute("innerHTML"), 'html.parser')
                        dict_spec.update(self._soup_to_dict(soup))
                    ActionChains(driver).key_down(Keys.PAGE_DOWN).perform()
                if self.tracking_log:
                    driver.save_screenshot(f"./{dir_model}/{stamp_url}_4_end_{stamp_today}.png")
                if self.tracking_log:
                    print(f"Received information from {url}")
                return dict_spec
            except Exception as e:
                if self.tracking_log:
                    if cnt_try + 1 == try_total:
                        print(f"An error occurred on page 3rd : {model}")

            finally:
                driver.quit()
                
    def _extract_model_info(self, model):
        """
        Extract additional information from the model name.
        """
        dict_info = {}
        dict_info["year"] = model.split("-")[1][-1]
        dict_info["series"] = model.split("-")[1][2:]
        dict_info["size"] = model.split("-")[1][:2]
        dict_info["grade"] = model.split("-")[0]

        year_mapping = {
            'L': "2023",
            'K': "2022",
            'J': "2021",
            # Add additional mappings as needed
        }

        try:
            dict_info["year"] = year_mapping.get(dict_info.get("year"))
        except:
            dict_info["year"] = None

        return dict_info

    def _soup_to_dict(self, soup):
        """
        Convert BeautifulSoup soup to dictionary.
        """
        try:
            h4_tag = soup.find('h4').text.strip()
            p_tag = soup.find('p').text.strip()
        except :
            try:
                h4_tag = soup.find('h4').text.strip()
                p_tag = ""
            except Exception as e:
                print("Parser error", e)
                h4_tag = "Parser error"
                p_tag = "Parser error"
        return {h4_tag: p_tag}